@@ -2,30 +2,45 @@
 Open compressed files transparently.
 """
 
-__all__ = ["xopen", "PipedGzipWriter", "PipedGzipReader", "__version__"]
+__all__ = [
+    "xopen",
+    "PipedGzipReader",
+    "PipedGzipWriter",
+    "PipedIGzipReader",
+    "PipedIGzipWriter",
+    "PipedPigzReader",
+    "PipedPigzWriter",
+    "PipedPBzip2Reader",
+    "PipedPBzip2Writer",
+    "PipedPythonIsalReader",
+    "PipedPythonIsalWriter",
+    "__version__",
+]
 
 import gzip
 import sys
 import io
 import os
 import bz2
-import time
+import lzma
 import stat
 import signal
 import pathlib
 import subprocess
 import tempfile
+import time
 from abc import ABC, abstractmethod
 from subprocess import Popen, PIPE, DEVNULL
-from typing import Optional, TextIO, AnyStr, IO
+from typing import Optional, TextIO, AnyStr, IO, List, Set
 
 from ._version import version as __version__
 
 
 try:
-    import lzma
+    from isal import igzip, isal_zlib  # type: ignore
 except ImportError:
-    lzma = None  # type: ignore
+    igzip = None
+    isal_zlib = None
 
 try:
     import fcntl
@@ -38,25 +53,10 @@ except ImportError:
     fcntl = None  # type: ignore
 
 _MAX_PIPE_SIZE_PATH = pathlib.Path("/proc/sys/fs/pipe-max-size")
-if _MAX_PIPE_SIZE_PATH.exists():
-    _MAX_PIPE_SIZE = int(_MAX_PIPE_SIZE_PATH.read_text())  # type: Optional[int]
-else:
-    _MAX_PIPE_SIZE = None
-
-
 try:
-    from os import fspath  # Exists in Python 3.6+
-except ImportError:
-    def fspath(path):  # type: ignore
-        if hasattr(path, "__fspath__"):
-            return path.__fspath__()
-        # Python 3.4 and 3.5 have pathlib, but do not support the file system
-        # path protocol
-        if pathlib is not None and isinstance(path, pathlib.Path):
-            return str(path)
-        if not isinstance(path, str):
-            raise TypeError("path must be a string")
-        return path
+    _MAX_PIPE_SIZE = int(_MAX_PIPE_SIZE_PATH.read_text())  # type: Optional[int]
+except OSError:  # Catches file not found and permission errors. Possible other errors too.
+    _MAX_PIPE_SIZE = None
 
 
 def _available_cpu_count() -> int:
@@ -142,16 +142,16 @@ class Closing(ABC):
 
     @abstractmethod
     def close(self):
-        pass
+        """Called when exiting the context manager"""
 
 
 class PipedCompressionWriter(Closing):
     """
     Write Compressed files by running an external process and piping into it.
     """
-    def __init__(self, path, program: str, mode='wt',
+    def __init__(self, path, program_args: List[str], mode='wt',
                  compresslevel: Optional[int] = None,
-                 threads_flag: str = None,
+                 threads_flag: Optional[str] = None,
                  threads: Optional[int] = None):
         """
         mode -- one of 'w', 'wt', 'wb', 'a', 'at', 'ab'
@@ -168,11 +168,11 @@ class PipedCompressionWriter(Closing):
 
         # TODO use a context manager
         self.outfile = open(path, mode)
-        self.closed = False
-        self.name = path
-        self._mode = mode
-        self._program = program
-        self._threads_flag = threads_flag
+        self.closed: bool = False
+        self.name: str = path
+        self._mode: str = mode
+        self._program_args: List[str] = program_args
+        self._threads_flag: Optional[str] = threads_flag
 
         if threads is None:
             threads = min(_available_cpu_count(), 4)
@@ -196,16 +196,16 @@ class PipedCompressionWriter(Closing):
             self.__class__.__name__,
             self.name,
             self._mode,
-            self._program,
+            " ".join(self._program_args),
             self._threads,
         )
 
     def _open_process(
         self, mode: str, compresslevel: Optional[int], threads: int, outfile: TextIO,
     ) -> Popen:
-        program_args = [self._program]
+        program_args: List[str] = self._program_args[:]  # prevent list aliasing
         if threads != 0 and self._threads_flag is not None:
-            program_args += [self._threads_flag, str(threads)]
+            program_args += [f"{self._threads_flag}{threads}"]
         extra_args = []
         if 'w' in mode and compresslevel is not None:
             extra_args += ['-' + str(compresslevel)]
@@ -234,7 +234,8 @@ class PipedCompressionWriter(Closing):
         self.outfile.close()
         if retcode != 0:
             raise OSError(
-                "Output {} process terminated with exit code {}".format(self._program, retcode))
+                "Output {} process terminated with exit code {}".format(
+                    " ".join(self._program_args), retcode))
 
     def __iter__(self):  # type: ignore
         # For compatibility with Pandas, which checks for an __iter__ method
@@ -250,10 +251,16 @@ class PipedCompressionReader(Closing):
     Open a pipe to a process for reading a compressed file.
     """
 
+    # This exit code is not interpreted as an error when terminating the process
+    _allowed_exit_code: Optional[int] = -signal.SIGTERM
+    # If this message is printed on stderr on terminating the process,
+    # it is not interpreted as an error
+    _allowed_exit_message: Optional[bytes] = None
+
     def __init__(
         self,
         path,
-        program: str,
+        program_args: List[str],
         mode: str = "r",
         threads_flag: Optional[str] = None,
         threads: Optional[int] = None,
@@ -263,8 +270,8 @@ class PipedCompressionReader(Closing):
         """
         if mode not in ('r', 'rt', 'rb'):
             raise ValueError("Mode is '{}', but it must be 'r', 'rt' or 'rb'".format(mode))
-        self._program = program
-        program_args = [program, '-cd', path]
+        self._program_args = program_args
+        program_args = program_args + ['-cd', path]
 
         if threads_flag is not None:
             if threads is None:
@@ -275,7 +282,7 @@ class PipedCompressionReader(Closing):
                 #   using multiple threads while there is only a 10% gain in wall
                 #   clock time.
                 threads = 1
-            program_args += [threads_flag, str(threads)]
+            program_args += [f"{threads_flag}{threads}"]
         self._threads = threads
         self.process = Popen(program_args, stdout=PIPE, stderr=PIPE)
         self.name = path
@@ -285,15 +292,11 @@ class PipedCompressionReader(Closing):
 
         self._mode = mode
         if 'b' not in mode:
-            self._file = io.TextIOWrapper(self.process.stdout)  # type: IO
+            self._file: IO = io.TextIOWrapper(self.process.stdout)
         else:
             self._file = self.process.stdout
-        assert self.process.stderr is not None
-        self._stderr = io.TextIOWrapper(self.process.stderr)
         self.closed = False
-        # Give the subprocess a little bit of time to report any errors (such as
-        # a non-existing file)
-        time.sleep(0.01)
+        self._wait_for_output_or_process_exit()
         self._raise_if_error()
 
     def __repr__(self):
@@ -301,7 +304,7 @@ class PipedCompressionReader(Closing):
             self.__class__.__name__,
             self.name,
             self._mode,
-            self._program,
+            " ".join(self._program_args),
             self._threads,
         )
 
@@ -310,16 +313,14 @@ class PipedCompressionReader(Closing):
             return
         self.closed = True
         retcode = self.process.poll()
+        check_allowed_code_and_message = False
         if retcode is None:
             # still running
             self.process.terminate()
-            allow_sigterm = True
-        else:
-            allow_sigterm = False
-        self.process.wait()
+            check_allowed_code_and_message = True
+        _, stderr_message = self.process.communicate()
         self._file.close()
-        self._raise_if_error(allow_sigterm=allow_sigterm)
-        self._stderr.close()
+        self._raise_if_error(check_allowed_code_and_message, stderr_message)
 
     def __iter__(self):
         return self
@@ -327,21 +328,60 @@ class PipedCompressionReader(Closing):
     def __next__(self) -> AnyStr:
         return self._file.__next__()
 
-    def _raise_if_error(self, allow_sigterm: bool = False) -> None:
+    def _wait_for_output_or_process_exit(self):
         """
-        Raise IOError if process is not running anymore and the exit code is
-        nonzero. If allow_sigterm is set and a SIGTERM exit code is
-        encountered, no error is raised.
+        Wait for the process to produce at least some output, or has exited.
+        """
+        # The program may crash due to a non-existing file, internal error etc.
+        # In that case we need to check. However the 'time-to-crash' differs
+        # between programs. Some crash faster than others.
+        # Therefore we peek the first character(s) of stdout. Peek will return at
+        # least one byte of data, unless the buffer is empty or at EOF. If at EOF,
+        # we should wait for the program to exit. This way we ensure the program
+        # has at least decompressed some output, or stopped before we continue.
+
+        # stdout is io.BufferedReader if set to PIPE
+        while True:
+            first_output = self.process.stdout.peek(1)  # type: ignore
+            if first_output or self.process.poll() is not None:
+                break
+            time.sleep(0.01)
+
+    def _raise_if_error(self, check_allowed_code_and_message: bool = False,
+                        stderr_message: bytes = b"") -> None:
+        """
+        Raise OSError if process is not running anymore and the exit code is
+        nonzero. If check_allowed_code_and_message is set, OSError is not raised when
+        (1) the exit value of the process is equal to the value of the allowed_exit_code
+        attribute or (2) the allowed_exit_message attribute is set and it matches with
+        stderr_message.
         """
         retcode = self.process.poll()
-        if (
-            retcode is not None and retcode != 0
-            and not (allow_sigterm and retcode == -signal.SIGTERM)
-        ):
-            message = self._stderr.read().strip()
-            self._file.close()
-            self._stderr.close()
-            raise OSError("{} (exit code {})".format(message, retcode))
+
+        if retcode is None:
+            # process still running
+            return
+        if retcode == 0:
+            # process terminated successfully
+            return
+
+        if check_allowed_code_and_message:
+            if retcode == self._allowed_exit_code:
+                # terminated with allowed exit code
+                return
+            if (
+                self._allowed_exit_message
+                and stderr_message.startswith(self._allowed_exit_message)
+            ):
+                # terminated with another exit code, but message is allowed
+                return
+
+        assert self.process.stderr is not None
+        if not stderr_message:
+            stderr_message = self.process.stderr.read()
+
+        self._file.close()
+        raise OSError("{!r} (exit code {})".format(stderr_message, retcode))
 
     def read(self, *args) -> AnyStr:
         return self._file.read(*args)
@@ -356,7 +396,10 @@ class PipedCompressionReader(Closing):
         return self._file.seekable()
 
     def peek(self, n: int = None):
-        return self._file.peek(n)  # type: ignore
+        if hasattr(self._file, "peek"):
+            return self._file.peek(n)  # type: ignore
+        else:
+            raise AttributeError("Peek is not available when 'b' not in mode")
 
     def readable(self) -> bool:
         return self._file.readable()
@@ -369,6 +412,34 @@ class PipedCompressionReader(Closing):
 
 
 class PipedGzipReader(PipedCompressionReader):
+    """
+    Open a pipe to gzip for reading a gzipped file.
+    """
+    def __init__(self, path, mode: str = "r"):
+        super().__init__(path, ["gzip"], mode)
+
+
+class PipedGzipWriter(PipedCompressionWriter):
+    """
+    Write gzip-compressed files by running an external gzip process and
+    piping into it. On Python 3, gzip.GzipFile is on par with gzip itself,
+    but running an external gzip can still reduce wall-clock time because
+    the compression happens in a separate process.
+    """
+    def __init__(self, path, mode: str = "wt", compresslevel: Optional[int] = None):
+        """
+        mode -- one of 'w', 'wt', 'wb', 'a', 'at', 'ab'
+        compresslevel -- compression level
+        threads (int) -- number of pigz threads. If this is set to None, a reasonable default is
+            used. At the moment, this means that the number of available CPU cores is used, capped
+            at four to avoid creating too many threads. Use 0 to let pigz use all available cores.
+        """
+        if compresslevel is not None and compresslevel not in range(1, 10):
+            raise ValueError("compresslevel must be between 1 and 9")
+        super().__init__(path, ["gzip"], mode, compresslevel, None)
+
+
+class PipedPigzReader(PipedCompressionReader):
     """
     Open a pipe to pigz for reading a gzipped file. Even though pigz is mostly
     used to speed up writing by using many compression threads, it is
@@ -376,21 +447,18 @@ class PipedGzipReader(PipedCompressionReader):
     (ca. 2x speedup).
     """
     def __init__(self, path, mode: str = "r", threads: Optional[int] = None):
-        try:
-            super().__init__(path, "pigz", mode, "-p", threads)
-        except OSError:
-            super().__init__(path, "gzip", mode, None, threads)
+        super().__init__(path, ["pigz"], mode, "-p", threads)
 
 
-class PipedGzipWriter(PipedCompressionWriter):
+class PipedPigzWriter(PipedCompressionWriter):
     """
-    Write gzip-compressed files by running an external gzip or pigz process and
-    piping into it. pigz is tried first. It is fast because it can compress using
-    multiple cores. Also it is more efficient on one core.
-    If pigz is not available, a gzip subprocess is used. On Python 3, gzip.GzipFile is on
-    par with gzip itself, but running an external gzip can still reduce wall-clock
-    time because the compression happens in a separate process.
+    Write gzip-compressed files by running an external pigz process and
+    piping into it. pigz can compress using multiple cores. It is also more
+    efficient than gzip on only one core. (But then igzip is even faster and
+    should be preferred if the compression level allows it.)
     """
+    _accepted_compression_levels: Set[int] = set(list(range(10)) + [11])
+
     def __init__(
         self,
         path,
@@ -405,12 +473,37 @@ class PipedGzipWriter(PipedCompressionWriter):
             used. At the moment, this means that the number of available CPU cores is used, capped
             at four to avoid creating too many threads. Use 0 to let pigz use all available cores.
         """
-        if compresslevel is not None and compresslevel not in range(1, 10):
-            raise ValueError("compresslevel must be between 1 and 9")
-        try:
-            super().__init__(path, "pigz", mode, compresslevel, "-p", threads)
-        except OSError:
-            super().__init__(path, "gzip", mode, compresslevel, None, threads)
+        if compresslevel is not None and compresslevel not in self._accepted_compression_levels:
+            raise ValueError("compresslevel must be between 0 and 9 or 11")
+        super().__init__(path, ["pigz"], mode, compresslevel, "-p", threads)
+
+
+class PipedPBzip2Reader(PipedCompressionReader):
+    """
+    Open a pipe to pbzip2 for reading a bzipped file.
+    """
+
+    _allowed_exit_code = None
+    _allowed_exit_message = b"\n *Control-C or similar caught [sig=15], quitting..."
+
+    def __init__(self, path, mode: str = "r", threads: Optional[int] = None):
+        super().__init__(path, ["pbzip2"], mode, "-p", threads)
+
+
+class PipedPBzip2Writer(PipedCompressionWriter):
+    """
+    Write bzip2-compressed files by running an external pbzip2 process and
+    piping into it. pbzip2 can compress using multiple cores.
+    """
+
+    def __init__(
+        self,
+        path,
+        mode: str = "wt",
+        threads: Optional[int] = None,
+    ):
+        # Use default compression level for pbzip2: 9
+        super().__init__(path, ["pbzip2"], mode, 9, "-p", threads)
 
 
 class PipedIGzipReader(PipedCompressionReader):
@@ -429,7 +522,7 @@ class PipedIGzipReader(PipedCompressionReader):
                 "This version of igzip does not support reading "
                 "concatenated gzip files and is therefore not "
                 "safe to use. See: https://github.com/intel/isa-l/issues/143")
-        super().__init__(path, "igzip", mode)
+        super().__init__(path, ["igzip"], mode)
 
 
 class PipedIGzipWriter(PipedCompressionWriter):
@@ -449,7 +542,19 @@ class PipedIGzipWriter(PipedCompressionWriter):
     def __init__(self, path, mode: str = "wt", compresslevel: Optional[int] = None):
         if compresslevel is not None and compresslevel not in range(0, 4):
             raise ValueError("compresslevel must be between 0 and 3")
-        super().__init__(path, "igzip", mode, compresslevel)
+        super().__init__(path, ["igzip"], mode, compresslevel)
+
+
+class PipedPythonIsalReader(PipedCompressionReader):
+    def __init__(self, path, mode: str = "r"):
+        super().__init__(path, [sys.executable, "-m", "isal.igzip"], mode)
+
+
+class PipedPythonIsalWriter(PipedCompressionWriter):
+    def __init__(self, path, mode: str = "wt", compresslevel: Optional[int] = None):
+        if compresslevel is not None and compresslevel not in range(0, 4):
+            raise ValueError("compresslevel must be between 0 and 3")
+        super().__init__(path, [sys.executable, "-m", "isal.igzip"], mode, compresslevel)
 
 
 def _open_stdin_or_out(mode: str) -> IO:
@@ -459,42 +564,84 @@ def _open_stdin_or_out(mode: str) -> IO:
     return open(std.fileno(), mode=mode, closefd=False)
 
 
-def _open_bz2(filename, mode: str) -> IO:
+def _open_bz2(filename, mode: str, threads: Optional[int]):
+    if threads != 0:
+        try:
+            if "r" in mode:
+                return PipedPBzip2Reader(filename, mode, threads)
+            else:
+                return PipedPBzip2Writer(filename, mode, threads)
+        except OSError:
+            pass  # We try without threads.
+
     return bz2.open(filename, mode)
 
 
 def _open_xz(filename, mode: str) -> IO:
-    if lzma is None:
-        raise ImportError(
-            "Cannot open xz files: The lzma module is not available (use Python 3.3 or newer)")
     return lzma.open(filename, mode)
 
 
+def _open_external_gzip_reader(filename, mode, compresslevel, threads):
+    assert "r" in mode
+    try:
+        return PipedIGzipReader(filename, mode)
+    except (OSError, ValueError):
+        # No igzip installed or version does not support reading
+        # concatenated files.
+        pass
+    if igzip:
+        return PipedPythonIsalReader(filename, mode)
+    try:
+        return PipedPigzReader(filename, mode, threads=threads)
+    except OSError:
+        return PipedGzipReader(filename, mode)
+
+
+def _open_external_gzip_writer(filename, mode, compresslevel, threads):
+    assert "r" not in mode
+    try:
+        return PipedIGzipWriter(filename, mode, compresslevel)
+    except (OSError, ValueError):
+        # No igzip installed or compression level higher than 3
+        pass
+    if igzip:  # We can use the CLI from isal.igzip
+        try:
+            return PipedPythonIsalWriter(filename, mode, compresslevel)
+        except ValueError:  # Wrong compression level
+            pass
+    try:
+        return PipedPigzWriter(filename, mode, compresslevel, threads=threads)
+    except OSError:
+        return PipedGzipWriter(filename, mode, compresslevel)
+
+
 def _open_gz(filename, mode: str, compresslevel, threads):
     if threads != 0:
         try:
-            if 'r' in mode:
-                try:
-                    return PipedIGzipReader(filename, mode)
-                except (OSError, ValueError):
-                    # No igzip installed or version does not support reading
-                    # concatenated files.
-                    return PipedGzipReader(filename, mode, threads=threads)
+            if "r" in mode:
+                return _open_external_gzip_reader(filename, mode, compresslevel, threads)
             else:
-                try:
-                    return PipedIGzipWriter(filename, mode, compresslevel)
-                except (OSError, ValueError):
-                    # No igzip installed or compression level higher than 3
-                    return PipedGzipWriter(filename, mode, compresslevel, threads=threads)
+                return _open_external_gzip_writer(filename, mode, compresslevel, threads)
         except OSError:
             pass  # We try without threads.
 
     if 'r' in mode:
+        if igzip is not None:
+            return igzip.open(filename, mode)
         return gzip.open(filename, mode)
-    else:
-        # Override gzip.open's default of 9 for consistency with command-line gzip.
-        return gzip.open(filename, mode,
-                         compresslevel=6 if compresslevel is None else compresslevel)
+
+    if igzip is not None:
+        try:
+            return igzip.open(filename, mode,
+                              compresslevel=isal_zlib.ISAL_DEFAULT_COMPRESSION
+                              if compresslevel is None else compresslevel)
+        except ValueError:
+            # Compression level not supported, move to built-in gzip.
+            pass
+
+    # Override gzip.open's default of 9 for consistency with command-line gzip.
+    return gzip.open(filename, mode,
+                     compresslevel=6 if compresslevel is None else compresslevel)
 
 
 def _detect_format_from_content(filename: str) -> Optional[str]:
@@ -548,17 +695,18 @@ def xopen(
     bzip2 and xz. If the filename is '-', standard output (mode 'w') or
     standard input (mode 'r') is returned.
 
-    The file type is determined based on the filename: .gz is gzip, .bz2 is bzip2, .xz is
-    xz/lzma and no compression assumed otherwise.
+    When writing, the file format is chosen based on the file name extension:
+    - .gz uses gzip compression
+    - .bz2 uses bzip2 compression
+    - .xz uses xz/lzma compression
+    - otherwise, no compression is used
+
+    When reading, if a file name extension is available, the format is detected
+    using it, but if not, the format is detected from the contents.
 
     mode can be: 'rt', 'rb', 'at', 'ab', 'wt', or 'wb'. Also, the 't' can be omitted,
     so instead of 'rt', 'wt' and 'at', the abbreviations 'r', 'w' and 'a' can be used.
 
-    In Python 2, the 't' and 'b' characters are ignored.
-
-    Append mode ('a', 'at', 'ab') is not available with BZ2 compression and
-    will raise an error.
-
     compresslevel is the compression level for writing to gzip files.
     This parameter is ignored for the other compression formats. If set to
     None (default), level 6 is used.
@@ -574,7 +722,7 @@ def xopen(
         mode += 't'
     if mode not in ('rt', 'rb', 'wt', 'wb', 'at', 'ab'):
         raise ValueError("Mode '{}' not supported".format(mode))
-    filename = fspath(filename)
+    filename = os.fspath(filename)
 
     if filename == '-':
         return _open_stdin_or_out(mode)
@@ -588,6 +736,6 @@ def xopen(
     elif detected_format == "xz":
         return _open_xz(filename, mode)
     elif detected_format == "bz2":
-        return _open_bz2(filename, mode)
+        return _open_bz2(filename, mode, threads)
     else:
         return open(filename, mode)
